/*------------------------------------------------------------------------------------------*\
 *   Copyright (C) 2011 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 \*------------------------------------------------------------------------------------------*/

#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/types.h>
#include <linux/ethtool.h>
#include <linux/avm_hw_config.h>

#include "avmnet_debug.h"
#include "avmnet_config.h"
#include "avmnet_module.h"
#include "avmnet_common.h"
#include "avmnet_ar803x.h"
#include "athrs_phy_ctrl.h"
#include "mdio_reg.h"

#if defined(CONFIG_MACH_ATHEROS)
#include <atheros.h>
#include <asm/mach-atheros/atheros_gpio.h>
#endif

#define MODULE_NAME "AVMNET_AR8033"

#define AR803x_PHY_ID1 0x004D
#define AR803x_PHY_ID2 0xD072
#define AR8326_PHY_ID2 0xD042
#define AR8033_PHY_ID2 0xD043


#define SUSPEND_PARENT(this) do{ \
                                 if(this->parent->suspend){ \
                                     this->parent->suspend(this->parent, this); \
                                 } \
                             }while(0);

#define RESUME_PARENT(this) do{ \
                                if(this->parent->resume){ \
                                    this->parent->resume(this->parent, this); \
                                } \
                            }while(0);

#if defined(CONFIG_MACH_ATHEROS)
static struct resource ar8033_gpio ={.name = "ar8033_reset",
                                     .flags = IORESOURCE_IO,
                                     .start = -1,
                                     .end = -1};
#endif // CONFIG_MACH_ATHEROS != n
static int set_config(avmnet_module_t *this);
static int setup_phy(avmnet_module_t *this);
static void reset_phy(avmnet_module_t *this, enum rst_type type);
static avmnet_linkstatus_t get_status(avmnet_module_t *this);

static int phy_reg_open(struct inode *inode, struct file *file);
static int phy_reg_show(struct seq_file *seq, void *data __attribute__ ((unused)));
static ssize_t phy_reg_write(struct file *filp, const char __user *buff, size_t len,
        loff_t *offset);

static const struct file_operations phy_reg_fops = {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 32)
        .owner = THIS_MODULE,
#endif
        .open = phy_reg_open,
        .read = seq_read,
        .write = phy_reg_write,
        .llseek = seq_lseek,
        .release = seq_release, };

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static void link_work_handler(struct work_struct *data)
{
    struct athphy_context *phy;
    avmnet_module_t *this;
    avmnet_module_t *parent;
    avmnet_linkstatus_t status;

    phy = container_of(
            container_of(to_delayed_work(data), avmnet_work_t, work),
            struct athphy_context, link_work);

    this = (avmnet_module_t *) phy->this_module;
    parent = this->parent;

//    AVMNET_ERR("[%s] Called for module %s\n", __func__, this->name);

    status = get_status(this);
    parent->set_status(parent, this->device_id, status);

}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
#define LED_OVERRIDE 0x19
#define LED_NORMAL  0
#define LED_ACT     1
#define LED_OFF     3
#define LED_ON      2

union ar8033_led_override {
    unsigned short Register;
    struct _ar8033_led_override {
        unsigned int res13_15   : 3;
        unsigned int act_ctrl   : 1;
        unsigned int res8_11    : 4;
        unsigned int led_link   : 2;
        unsigned int res45      : 2;
        unsigned int led_tx     : 2;
        unsigned int led_rx     : 2;
    } Bits;
};

static void gpio_work_handler(struct work_struct *data)
{
    struct athphy_context *phy;
    avmnet_module_t *this;
    unsigned int gpio;
    unsigned int on;
    union ar8033_led_override value;

    phy = container_of(
            container_of(to_delayed_work(data), avmnet_work_t, work),
            struct athphy_context, gpio_work);

    this = (avmnet_module_t *) phy->this_module;


    if(phy->powerdown){
        return;
    }

    if(down_interruptible(&(phy->mutex))){
        return;
    }

    AVMNET_DEBUG("(%s) read %d write %d\n", __func__, phy->gpio.read, phy->gpio.write);

    while (phy->gpio.read != phy->gpio.write) {

        value.Register = MDIO_READ(LED_OVERRIDE);
        AVMNET_DEBUG("(%s) LED_OVERRIDE 0x%x read %d gpio 0x%x\n", __func__, 
                value.Register, phy->gpio.read, phy->gpio.queue[phy->gpio.read]);

        gpio = phy->gpio.queue[phy->gpio.read] & 0xff;
        on   = phy->gpio.queue[phy->gpio.read] >> 8;
        phy->gpio.read++;
        if (phy->gpio.read == AVMNET_GPIO_QUEUELEN)
            phy->gpio.read = 0;

        if (gpio & (1<<0)) {
            if (on) {
                value.Bits.led_tx = LED_ON;
                value.Bits.led_rx = LED_ON;
            } else {
                value.Bits.led_tx = LED_OFF;
                value.Bits.led_rx = LED_OFF;
            }
        }

        if (gpio & (1<<1)) {
            if (on) {
                value.Bits.led_link = LED_ON;
            } else {
                value.Bits.led_link = LED_OFF;
            }
        }

        MDIO_WRITE(LED_OVERRIDE, value.Register);

        AVMNET_DEBUG("(%s) LED_OVERRIDE value 0x%x\n", __func__, value.Register);
    }

    up(&(phy->mutex));

}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8033_gpio_set(avmnet_module_t *this, unsigned int gpio, unsigned int on_off) {

    struct athphy_context *phy = (struct athphy_context *) this->priv;

    AVMNET_DEBUG("[%s] device %s gpio %d on_off %d\n", __func__, this->name, gpio, on_off);

    phy->gpio.queue[phy->gpio.write++] = (1 << gpio) | (on_off << 8);
    if (phy->gpio.write == AVMNET_GPIO_QUEUELEN) 
        phy->gpio.write = 0;

    queue_delayed_work(phy->gpio_work.workqueue, &(phy->gpio_work.work), 0);

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static irqreturn_t ar8033_link_intr(int irq __attribute__((unused)), void *dev)
{
    avmnet_module_t *this = (avmnet_module_t *) dev;
    struct athphy_context *phy = (struct athphy_context *) this->priv;
    unsigned long flags;
    unsigned int irqStatus;

    spin_lock_irqsave(&(phy->irq_lock), flags);

//    AVMNET_DEBUG("[%s] device %s irq %d\n", __func__, this->name, phy->irq);

    irqStatus   = ath_reg_rd(SGMII_INTERRUPT_ADDRESS);

    AVMNET_DEBUG("[%s] device %s irq 0x%x mask 0x%x Debug 0x%x\n", __func__, 
            this->name, irqStatus, ath_reg_rd(SGMII_INTERRUPT_MASK_ADDRESS), ath_reg_rd(0x18070058));

    ath_reg_rmw_clear(SGMII_INTERRUPT_ADDRESS, irqStatus);     /*--- clear IRQ-Bits ---*/

    queue_delayed_work(phy->link_work.workqueue, &(phy->link_work.work), 0);

    spin_unlock_irqrestore(&(phy->irq_lock), flags);

    return IRQ_HANDLED;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static avmnet_linkstatus_t get_status(avmnet_module_t *this)
{
    struct athphy_context *phy = (struct athphy_context *) this->priv;
    avmnet_linkstatus_t status;
    uint16_t phyStatus, phyHwStatus, phyControl, phyResolve, irqStatus;

    status.Status = 0;

    if(phy->powerdown){
        return status;
    }

    if(down_interruptible(&(phy->mutex))){
        return status;
    }

    irqStatus   = ath_reg_rd(SGMII_INTERRUPT_ADDRESS);
    phyControl  = ath_reg_rd(MR_AN_CONTROL_ADDRESS);
    phyStatus   = ath_reg_rd(MR_AN_STATUS_ADDRESS);
    phyHwStatus = ath_reg_rd(SGMII_MAC_RX_CONFIG_ADDRESS);
    phyResolve  = ath_reg_rd(SGMII_RESOLVE_ADDRESS);

    AVMNET_DEBUG("[%s] device %s irqStatus 0x%x phyControl 0x%x phyStatus 0x%x phyHwStatus 0x%x\n", 
            __func__, this->name, irqStatus, phyControl, phyStatus, phyHwStatus);

#if 0
    printk(KERN_ERR "{%s} 0x18070014 SGMII_RESET 0x%x\n", __func__, ath_reg_rd(0x18070014));
    printk(KERN_ERR "{%s} 0x18070018 SGMII_SERDES 0x%x\n", __func__, ath_reg_rd(0x18070018));
    printk(KERN_ERR "{%s} 0x18070034 SGMII_CONFIG 0x%x\n", __func__, ath_reg_rd(0x18070034));
    printk(KERN_ERR "{%s} 0x18070038 SGMII_MAC_RX_CONFIG 0x%x\n", __func__, ath_reg_rd(0x18070038));
    printk(KERN_ERR "{%s} 0x1807003C SGMII_MAC_TX_CONFIG 0x%x\n", __func__, ath_reg_rd(0x1807003C));
    printk(KERN_ERR "{%s} 0x18070048 EEE 0x%x\n", __func__, ath_reg_rd(0x18070048));
    printk(KERN_ERR "{%s} 0x18070054 SGMII_RESOLVE 0x%x\n", __func__, ath_reg_rd(0x18070054));
    printk(KERN_ERR "{%s} 0x18070064 PRBS_STATUS 0x%x\n", __func__, ath_reg_rd(0x18070064));
#endif

    // first check that  phy is not in power-down mode
    if( ! MR_AN_CONTROL_POWER_DOWN_GET(phyControl)){
        status.Details.powerup = 1;

        if((phyStatus & (1<<2)) && 
           (phy->config.autoneg == AUTONEG_DISABLE || (phyStatus & (1<<5))))
        {
            AVMNET_DEBUG("[%s] device %s <PHY found LINK>\n", __func__, this->name);
            status.Details.link = 1;
            status.Details.fullduplex = SGMII_MAC_RX_CONFIG_DUPLEX_MODE_GET(phyHwStatus);
            status.Details.speed = SGMII_MAC_RX_CONFIG_SPEED_MODE_GET(phyHwStatus);

#if 0
            if(phyHwStatus & PHY_SPEC_STATUS_MDIX){
                status.Details.mdix = AVMNET_LINKSTATUS_MDI_MDIX;
            }else{
                status.Details.mdix = AVMNET_LINKSTATUS_MDI_MDI;
            }
#endif

            if( SGMII_RESOLVE_RX_PAUSE_GET(phyResolve)){
                status.Details.flowcontrol |= AVMNET_LINKSTATUS_FC_RX;
            }

            if( SGMII_RESOLVE_RX_PAUSE_GET(phyResolve) && ( phy->config.pause_setup != avmnet_pause_rx) ){

                // TX pause: there is no way to advertise RX-Only, so we advertise
                // symetric and asymetric RX, and disable TX pause frames here
                status.Details.flowcontrol |= AVMNET_LINKSTATUS_FC_TX;
            }

        } else {
            AVMNET_DEBUG("[%s] device %s <PHY no LINK>\n", __func__, this->name);
        }

    } else {
        AVMNET_DEBUG("[%s] device %s PHY in POWERDOWN\n", __func__, this->name);
    }

    up(&(phy->mutex));

#if 0
    {
        unsigned int i;
        for (i = 0; i < 0x20; i++) {
            printk("{%s} PHY-Register [%d] 0x%x\n", __func__, i, MDIO_READ(i));
        }
    }
#endif

    return status;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8033_status_poll(avmnet_module_t *this)
{
    avmnet_module_t *parent = this->parent;
    avmnet_linkstatus_t status;
    int i, result;

    status = get_status(this);

    if(!(this->device_id->flags & AVMNET_DEVICE_FLAG_INITIALIZED)){
        AVMNET_ERR("[%s] %s not initialized yet\n", __func__, this->device_id->device_name);
        dump_stack();
        return 0;
    }

    for(i = 0; i < this->num_children; ++i){
        result = this->children[i]->poll(this->children[i]);
        if(result < 0){
            AVMNET_WARN("Module %s: poll() failed on child %s\n", this->name, this->children[i]->name);
        }
    }

    parent->set_status(parent, this->device_id, status);

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int setup_interrupt(avmnet_module_t *this, unsigned int on_off)
{
    struct athphy_context *phy = (struct athphy_context *) this->priv;

    AVMNET_DEBUG("[%s] %sable PHY-Interrupt on %s\n", __FUNCTION__, on_off ? "en":"dis", this->name);

    // warn on unsafe call
    if(!down_trylock(&(phy->mutex))){
        AVMNET_ERR("[%s] Called without MUTEX held!\n", __func__);
        dump_stack();
        up(&(phy->mutex));
    }

    if(this->initdata.phy.flags & AVMNET_CONFIG_FLAG_IRQ) {
        if(on_off){
            ath_reg_wr(SGMII_INTERRUPT_MASK_ADDRESS, SGMII_INTR);
        }else{
            ath_reg_wr(SGMII_INTERRUPT_MASK_ADDRESS, 0);
        }

        phy->irq_on = on_off;

        return 0;
    }else{
        return 1;
    }
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8033_setup_interrupt(avmnet_module_t *this, unsigned int on_off)
{
    struct athphy_context *phy = (struct athphy_context *) this->priv;
    int result;

    if(down_interruptible(&(phy->mutex))){
        return -EINTR;
    }

    result = setup_interrupt(this, on_off);

    up(&(phy->mutex));

    return result;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8033_setup(avmnet_module_t *this)
{
    struct athphy_context *phy = (struct athphy_context *) this->priv;
    int i, result;

    AVMNET_INFO("[%s] Called for PHY %s\n", __func__, this->name);

    if(down_interruptible(&(phy->mutex))){
        return -EINTR;
    }

    // default configuration
    if(this->initdata.phy.flags & AVMNET_CONFIG_FLAG_PHY_GBIT){
        phy->gbit = 1;
    }

    if(this->initdata.phy.flags & AVMNET_CONFIG_FLAG_PHY_MODE_CONF){
        phy->config.supported = this->initdata.phy.mode_conf;

        if((phy->config.supported & (ADVERTISED_1000baseT_Full | ADVERTISED_1000baseT_Half)) && phy->gbit == 0)
        {
            AVMNET_ERR("[%s] Config error for PHY %s: GBit modes configured on FE PHY.\n", __func__, this->name);
            phy->config.supported &= ~(ADVERTISED_1000baseT_Full | ADVERTISED_1000baseT_Half);
        }

    }else{
        phy->config.supported = ( ADVERTISED_TP
                                | ADVERTISED_Autoneg
                                | ADVERTISED_10baseT_Half
                                | ADVERTISED_10baseT_Full
                                | ADVERTISED_100baseT_Half
                                | ADVERTISED_100baseT_Full);

        if(phy->gbit){
            phy->config.supported |= (ADVERTISED_1000baseT_Full | ADVERTISED_1000baseT_Half);
        } 
    }

    // setup default ethtool flocwcontrol, based on pyh initdata
    if ( this->initdata.phy.flags & AVMNET_CONFIG_FLAG_DEFAULT_EN_FC_RX ){
        if ( this->initdata.phy.flags & AVMNET_CONFIG_FLAG_DEFAULT_EN_FC_TX ){
            phy->config.pause_setup = avmnet_pause_rx_tx;
        }
        else {
            // most reasonable default config for ethernet switches:
            // receive pause frames, but not to send pause frames
            phy->config.pause_setup = avmnet_pause_rx;
        }
    }
    else if ( this->initdata.phy.flags & AVMNET_CONFIG_FLAG_DEFAULT_EN_FC_TX ){
        phy->config.pause_setup = avmnet_pause_tx;
    }
    else {
        phy->config.pause_setup = avmnet_pause_none;
    }

    // what kind of flow control is supported by the MAC, we are attached to?
    phy->config.supported |= calc_advertise_pause_param(
            (this->parent->initdata.mac.flags &  AVMNET_CONFIG_FLAG_MAC_SUPPORT_FC_SYM),
            (this->parent->initdata.mac.flags &  AVMNET_CONFIG_FLAG_MAC_SUPPORT_FC_ASYM),
            phy->config.pause_setup );

    AVMNET_INFO("[%s] default pause config is %d, asym_pause_adv=%d, sym_pause_adv=%d.\n", __func__,phy->config.pause_setup,
            (phy->config.supported  &  ADVERTISED_Asym_Pause),
            (phy->config.supported  &  ADVERTISED_Pause));

    if(phy->config.supported & ADVERTISED_Autoneg){
        phy->config.autoneg = AUTONEG_ENABLE;
    }else{
        phy->config.autoneg = AUTONEG_DISABLE;
    }

    if(phy->config.supported & (ADVERTISED_1000baseT_Full | ADVERTISED_1000baseT_Half)){
        phy->config.speed = SPEED_1000;
    }else if(phy->config.supported & (ADVERTISED_100baseT_Full | ADVERTISED_100baseT_Half)){
        phy->config.speed = SPEED_100;
    }else{
        phy->config.speed = SPEED_10;
    }

    phy->config.advertise = phy->config.supported;
    phy->config.port = PORT_TP;
    phy->config.mdi = ETH_TP_MDI;

    if(this->initdata.phy.flags & AVMNET_CONFIG_FLAG_IRQ_ON){
        phy->irq_on = 1;
    }

    /*
     * how long to wait before the transition from link_failed to link_down
     */
    if(this->initdata.phy.flags & AVMNET_CONFIG_FLAG_RST_ON_LINKFAIL){
        phy->linkdown_delay = max(phy->linkf_time, (unsigned long int) (LINK_DOWN_DELAY * HZ));
    }else{
        phy->linkdown_delay = phy->linkf_time;
    }

    spin_lock_init(&(phy->irq_lock));

    result = setup_phy(this);

    AVMNET_DEBUG("[%s] Called for module %s SGMII_CONFIG 0x%x\n", __func__, this->name, ath_reg_rd(SGMII_CONFIG_ADDRESS));

    up(&(phy->mutex));

    if(result != 0){
        return result;
    }

    if(this->initdata.phy.flags & AVMNET_CONFIG_FLAG_IRQ){
        result = request_irq(phy->irq,
                             ar8033_link_intr,
                             IRQF_SHARED | IRQF_TRIGGER_LOW,
                             "ar8033 link",
                             this);
        if(result < 0){
            AVMNET_ERR("[%s] <request irq ar8033 link %d failed>\n", __FUNCTION__, phy->irq);
            return result;
        }
    }

    for(i = 0; i < this->num_children; ++i){
        result = this->children[i]->setup(this->children[i]);
        if(result != 0){
            // handle error
        }
    }

    result = avmnet_cfg_register_module(this);
    if(result < 0){
        AVMNET_ERR("[%s] avmnet_cfg_register_module failed for module %s\n", __func__, this->name);
        return result;
    }


    result = avmnet_cfg_add_seq_procentry(this, "mdio_regs", &phy_reg_fops); // cannot be done before avmnet_cfg_register_module
    if(result != 0){
        AVMNET_ERR("[%s] avmnet_cfg_add_seq_procentry failed for module %s\n", __func__, this->name);
        return result;
    }

    this->powerdown(this);

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
/*--- #define ATHR_SGMII_FORCED ---*/
#define SGMII_LINK_WAR_MAX_TRY 10
static void athr_gmac_sgmii_setup(void) {

	uint32_t status = 0, count = 0;

#if defined(ATHR_SGMII_FORCED)
		ath_reg_wr(MR_AN_CONTROL_ADDRESS, MR_AN_CONTROL_SPEED_SEL1_SET(1) |
				                           MR_AN_CONTROL_PHY_RESET_SET(1) | 
                                           MR_AN_CONTROL_DUPLEX_MODE_SET(1));
		udelay(10);
		ath_reg_wr(SGMII_CONFIG_ADDRESS, SGMII_CONFIG_MODE_CTRL_SET(2) |
				                          SGMII_CONFIG_FORCE_SPEED_SET(1) | 
                                          SGMII_CONFIG_SPEED_SET(2));
#else
		ath_reg_wr(SGMII_CONFIG_ADDRESS, SGMII_CONFIG_MODE_CTRL_SET(2));
		ath_reg_wr(MR_AN_CONTROL_ADDRESS, MR_AN_CONTROL_AN_ENABLE_SET(1) |
				                          MR_AN_CONTROL_PHY_RESET_SET(1));
#endif

    /*------------------------------------------------------------------------------------------*\
	 * SGMII reset sequence suggested by systems team.
    \*------------------------------------------------------------------------------------------*/
	ath_reg_wr(SGMII_RESET_ADDRESS, SGMII_RESET_RX_CLK_N_RESET);
	ath_reg_wr(SGMII_RESET_ADDRESS, SGMII_RESET_HW_RX_125M_N_SET(1));
	ath_reg_wr(SGMII_RESET_ADDRESS, SGMII_RESET_HW_RX_125M_N_SET(1) |
			                         SGMII_RESET_RX_125M_N_SET(1));

	ath_reg_wr(SGMII_RESET_ADDRESS, SGMII_RESET_HW_RX_125M_N_SET(1) |
			                         SGMII_RESET_TX_125M_N_SET(1) |
			                         SGMII_RESET_RX_125M_N_SET(1));

	ath_reg_wr(SGMII_RESET_ADDRESS, SGMII_RESET_HW_RX_125M_N_SET(1) |
			                         SGMII_RESET_TX_125M_N_SET(1) |
			                         SGMII_RESET_RX_125M_N_SET(1) |
			                         SGMII_RESET_RX_CLK_N_SET(1));

	ath_reg_wr(SGMII_RESET_ADDRESS, SGMII_RESET_HW_RX_125M_N_SET(1) |
			                         SGMII_RESET_TX_125M_N_SET(1) |
			                         SGMII_RESET_RX_125M_N_SET(1) |
			                         SGMII_RESET_RX_CLK_N_SET(1) |
			                         SGMII_RESET_TX_CLK_N_SET(1));

	ath_reg_rmw_clear(MR_AN_CONTROL_ADDRESS, MR_AN_CONTROL_PHY_RESET_SET(1));

    /*------------------------------------------------------------------------------------------*\
     * WAR::Across resets SGMII link status goes to weird
	 * state.
	 * if 0xb8070058 (SGMII_DEBUG register) reads other then 0x1f or 0x10
	 * for sure we are in bad  state.
	 * Issue a PHY reset in MR_AN_CONTROL_ADDRESS to keep going.
    \*------------------------------------------------------------------------------------------*/
    do {
		ath_reg_rmw_set(MR_AN_CONTROL_ADDRESS, MR_AN_CONTROL_PHY_RESET_SET(1));
		udelay(200);
		ath_reg_rmw_clear(MR_AN_CONTROL_ADDRESS, MR_AN_CONTROL_PHY_RESET_SET(1));
        mdelay(1000);
		if (count++ == SGMII_LINK_WAR_MAX_TRY) {
			AVMNET_ERR("[%s] Max resets limit reached exiting...\n", __func__);
			break;
		}
		status = (ath_reg_rd(SGMII_DEBUG_ADDRESS) & 0xff);
    } while (!(status == 0xf || status == 0x10));

	AVMNET_INFO("[%s] Done\n", __func__);
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int setup_phy(avmnet_module_t *this)
{
    struct athphy_context *phy = (struct athphy_context *) this->priv;
    unsigned int i;
    int status = 0;

    AVMNET_INFO("[%s] called for PHY %s\n", __func__, this->name);

    // warn on unsafe call
    if(!down_trylock(&(phy->mutex))){
        AVMNET_ERR("[%s] Called without MUTEX held!\n", __func__);
        dump_stack();
        up(&(phy->mutex));
    }

    reset_phy(this, rst_pulse);

    i = 0;
    while(MDIO_READ(MDIO_PHY_ID1) == 0xFFFF && i < 10){
        ++i;
        AVMNET_WARN("[%s] PHY %s hanging, trying to reset it.\n", __func__, this->name);
        reset_phy(this, rst_pulse);
        mdelay(1000);
    }

    AVMNET_DEBUG("{%s} PHY ID 0x%x\n", __func__, MDIO_READ(MDIO_PHY_ID1));

    if((i >= 10) || ( ! MDIO_READ(MDIO_PHY_ID1))) {
        AVMNET_ERR("[%s] Giving up on hanging PHY %s ID 0x%x\n", __func__, this->name, MDIO_READ(MDIO_PHY_ID1));
        return -EFAULT;
    }

    /*--- select CoperPageRegister, Media Copper, SGMII-Interface ---*/
    MDIO_WRITE(ATHR_PHY_CHIP_CONFIG, PHY_CHIP_CONFIG_BT_BX_REG_SEL | PHY_CHIP_CONFIG_PRIORITY_SEL | PHY_CHIP_CONFIG_BASET_SGMII);
    mdelay(100);
    MDIO_WRITE(ATHR_PHY_CONTROL, PHY_CTRL_AUTONEGOTIATION_ENABLE | PHY_CTRL_SOFTWARE_RESET);

#if 0
    { 
        unsigned short dummy;
        // setup SGMII-TX Outputvoltage
        MDIO_WRITE(MDIO_MMD_CTRL, 0x7);
        MDIO_WRITE(MDIO_MMD_ADDR, 0x8011);
        MDIO_WRITE(MDIO_MMD_CTRL, 0x4007);
        dummy = MDIO_READ(MDIO_MMD_ADDR);
        AVMNET_ERR("[%s] SGMII Drive CTRL: %04x\n", __func__, dummy);
    }
#endif

    if(this->initdata.phy.flags & AVMNET_CONFIG_FLAG_RX_DELAY){
        MDIO_WRITE(ATHR_PHY_DEBUG_PORT_ADDRESS, 0);
        MDIO_WRITE(ATHR_PHY_DEBUG_PORT_DATA, this->initdata.phy.rx_delay);
    }
    if(this->initdata.phy.flags & AVMNET_CONFIG_FLAG_TX_DELAY){
        MDIO_WRITE(ATHR_PHY_DEBUG_PORT_ADDRESS, 5);
        MDIO_WRITE(ATHR_PHY_DEBUG_PORT_DATA, this->initdata.phy.tx_delay);
    }

    if(this->initdata.phy.flags & AVMNET_CONFIG_FLAG_IRQ) {
        status |= setup_interrupt(this, phy->irq_on);
    }

    phy->state = LINK_DOWN;

    status |= set_config(this);

    athr_gmac_sgmii_setup();

    if(phy->powerdown){
        this->powerdown(this);
    }

    return status;

}
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8033_init(avmnet_module_t *this)
{
    struct athphy_context *phy;
    int i, result;

    AVMNET_TRC("{%s} Init on module %s called.\n", __func__, this->name);

    phy = kzalloc(sizeof(struct athphy_context), GFP_KERNEL);
    if(phy == NULL){
        AVMNET_ERR("{%s} init of avmnet-module %s failed.\n", __func__, this->name);
        return -ENOMEM;
    }

    this->priv = phy;
    phy->this_module = this;

    init_MUTEX(&(phy->mutex));
    down(&(phy->mutex)); // should not be necessary, just to shut up reset_phy()

    if(this->initdata.phy.flags & AVMNET_CONFIG_FLAG_MDIOADDR){
        phy->phyAddr = this->initdata.phy.mdio_addr;
    }else{
        AVMNET_ERR("[%s] No MDIO address for PHY %s.\n", __func__, this->name);
        BUG();
    }
    if(this->initdata.phy.flags & AVMNET_CONFIG_FLAG_RESET){
        phy->reset = this->initdata.phy.reset;

        if(phy->reset < 100){
            /*--- Hängt nicht am Shift-Register => GPIO Resourcen anmelden ---*/
            ar8033_gpio.start = phy->reset;
            ar8033_gpio.end = phy->reset;
            if(request_resource(&gpio_resource, &ar8033_gpio)){
                AVMNET_ERR(KERN_ERR "[%s] ERROR request resource gpio %d\n", __FUNCTION__, phy->reset);
                kfree(phy);
                return -EIO;
            }
        }

        reset_phy(this, rst_on);
    } else {
        AVMNET_ERR("[%s] No RESET configured for PHY %s.\n", __func__, this->name);
        BUG();
    }

    if(this->initdata.phy.flags & AVMNET_CONFIG_FLAG_LINKFAIL_TIME){
        phy->linkf_time = this->initdata.phy.lnkf_time;
    }else{
        phy->linkf_time = HZ;
    }

    phy->thrash_time = jiffies;

    phy->link_work.workqueue = avmnet_get_workqueue();
    if(phy->link_work.workqueue == NULL){
        AVMNET_ERR("[%s] No global workqueue available!\n>", __FUNCTION__);
        kfree(phy);
        return -EFAULT;
    }
    INIT_DELAYED_WORK(&(phy->link_work.work), link_work_handler);

    if (this->set_gpio) {
        phy->gpio_work.workqueue = avmnet_get_workqueue();
        if(phy->gpio_work.workqueue == NULL){
            AVMNET_ERR("[%s] No global workqueue available!\n>", __FUNCTION__);
            kfree(phy);
            return -EFAULT;
        }
        INIT_DELAYED_WORK(&(phy->gpio_work.work), gpio_work_handler);
        phy->gpio.write = 0;
        phy->gpio.read = 0;
    }

    up(&(phy->mutex));

    if(this->initdata.phy.flags & AVMNET_CONFIG_FLAG_IRQ){
        phy->irq = this->initdata.phy.irq;
    }

    for(i = 0; i < this->num_children; ++i){
        result = this->children[i]->init(this->children[i]);
        if(result != 0){
            // handle error
        }
    }

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8033_exit(avmnet_module_t *this)
{

    int i, result = 0;
    struct athphy_context *phy = (struct athphy_context *) this->priv;

    AVMNET_TRC("{%s} Exit on module %s called.\n", __func__, this->name);

    for(i = 0; i < this->num_children; ++i){
        result = this->children[i]->exit(this->children[i]);
        if(result != 0){
            // handle error
        }
    }

    if(this->initdata.phy.flags & AVMNET_CONFIG_FLAG_IRQ){
        free_irq(phy->irq, ar8033_link_intr);
        cancel_delayed_work_sync(&(phy->link_work.work));
    }

    kfree(this->priv);

    return result;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8033_powerdown(avmnet_module_t *this)
{
    struct athphy_context *phy = (struct athphy_context *) this->priv;
    int i, result;

    AVMNET_TRC("[%s] Called for module %s\n", __func__, this->name);

    for(i = 0; i < this->num_children; ++i){
        result = this->children[i]->powerdown(this->children[i]);
        if(result < 0){
            AVMNET_WARN("Module %s: powerdown() failed on child %s\n", this->name, this->children[i]->name);
        }
    }

    if(down_interruptible(&(phy->mutex))){
        return -EINTR;
    }

    MDIO_WRITE(MDIO_CONTROL, (MDIO_CONTROL_POWERDOWN | MDIO_CONTROL_ISOLATE));
    phy->powerdown = 1;

    up(&(phy->mutex));

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8033_powerup(avmnet_module_t *this)
{
    struct athphy_context *phy = (struct athphy_context *) this->priv;
    int i, result;

    AVMNET_TRC("[%s] Called for module %s\n", __func__, this->name);

    if(down_interruptible(&(phy->mutex))){
        return -EINTR;
    }

    MDIO_WRITE(MDIO_CONTROL, MDIO_CONTROL_RESET | MDIO_CONTROL_AUTONEG_ENABLE);
    msleep(500);

    phy->powerdown = 0;

    result = 0;
    /*
     * On broken PHYs, do a full reset on power-up. Otherwise just restore its
     * configuration
     */
    if(this->initdata.phy.flags & AVMNET_CONFIG_FLAG_RST_ON_LINKFAIL){
        SUSPEND_PARENT(this);

        result = setup_phy(this);

        RESUME_PARENT(this);
    } else {
        result = set_config(this);
        phy->state = LINK_DOWN;
    }


    up(&(phy->mutex));

    for(i = 0; i < this->num_children; ++i){
        result = this->children[i]->powerup(this->children[i]);
        if(result < 0){
            AVMNET_WARN("Module %s: powerup() failed on child %s\n", this->name, this->children[i]->name);
        }
    }

    return result;
}
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8033_ethtool_get_settings(avmnet_module_t *this, struct ethtool_cmd *cmd)
{
    unsigned int phyCtrl, phyAdv, phyGCtrl;
    avmnet_linkstatus_t status;
    struct athphy_context *phy = (struct athphy_context *) this->priv;

    AVMNET_TRC("[%s] Called for module %s\n", __func__, this->name);

    status = get_status(this);

    if(down_interruptible(&(phy->mutex))){
        return -EINTR;
    }

    phyCtrl = MDIO_READ(MDIO_CONTROL);
    phyAdv = MDIO_READ(MDIO_AUTONEG_ADV);
    phyGCtrl = MDIO_READ(MDIO_MASTER_SLAVE_CTRL);

    up(&(phy->mutex));

    cmd->supported = phy->config.supported;

    /*
     * Due to unreliable register reads in power-down mode, we just return
     * the configuration that will be set on power-up for advertising and autoneg
     */
    if(phy->powerdown){
        cmd->advertising = phy->config.advertise;
        cmd->autoneg = phy->config.autoneg;
    }else{
        cmd->advertising = 0;

        if(phyCtrl & MDIO_CONTROL_AUTONEG_ENABLE){
            cmd->autoneg = AUTONEG_ENABLE;

            if(phyAdv & MDIO_AUTONEG_ADV_10BT_HDX){
                cmd->advertising |= ADVERTISED_10baseT_Half;
            }

            if(phyAdv & MDIO_AUTONEG_ADV_10BT_FDX){
                cmd->advertising |= ADVERTISED_10baseT_Full;
            }

            if(phyAdv & MDIO_AUTONEG_ADV_100BT_HDX){
                cmd->advertising |= ADVERTISED_100baseT_Half;
            }

            if(phyAdv & MDIO_AUTONEG_ADV_100BT_FDX){
                cmd->advertising |= ADVERTISED_100baseT_Full;
            }

            if(phyAdv & MDIO_AUTONEG_ADV_10BT_HDX){
                cmd->advertising |= ADVERTISED_10baseT_Half;
            }

            if(phyAdv & MDIO_AUTONEG_ADV_PAUSE_ASYM){
                cmd->advertising |= ADVERTISED_Asym_Pause;
            }

            if(phyAdv & MDIO_AUTONEG_ADV_PAUSE_SYM){
                cmd->advertising |= ADVERTISED_Pause;
            }

            cmd->advertising |= ADVERTISED_TP | ADVERTISED_Autoneg;
        }else{
            cmd->autoneg = AUTONEG_DISABLE;
        }

        if(phy->gbit){
            if(cmd->autoneg == AUTONEG_ENABLE){
                if(phyGCtrl & MDIO_MASTER_SLAVE_CTRL_ADV_HD){
                    cmd->advertising |= ADVERTISED_1000baseT_Half;
                }

                if(phyGCtrl & MDIO_MASTER_SLAVE_CTRL_ADV_FD){
                    cmd->advertising |= ADVERTISED_1000baseT_Full;
                }
            }
        }
    }

    cmd->port = PORT_TP;
    cmd->phy_address = this->initdata.phy.mdio_addr;

    if(status.Details.link){
        switch(status.Details.speed){
        case 0:
            cmd->speed = SPEED_10;
            break;
        case 1:
            cmd->speed = SPEED_100;
            break;
        case 2:
            cmd->speed = SPEED_1000;
            break;
        default:
            cmd->speed = -1;
            break;
        }

        if(status.Details.fullduplex){
            cmd->duplex = DUPLEX_FULL;
        }else{
            cmd->duplex = DUPLEX_HALF;
        }

        if(status.Details.mdix == AVMNET_LINKSTATUS_MDI_MDIX){
            cmd->eth_tp_mdix = ETH_TP_MDI_X;
        }else{
            cmd->eth_tp_mdix = ETH_TP_MDI;
        }
    }else{
        cmd->speed = -1;
        cmd->duplex = -1;
        cmd->eth_tp_mdix = ETH_TP_MDI_INVALID;
    }

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8033_ethtool_set_settings(avmnet_module_t *this, struct ethtool_cmd *cmd)
{
    struct athphy_context *phy = (struct athphy_context *) this->priv;
    int result;
    unsigned int dummy_adv;

    AVMNET_DEBUG("[%s] Called for module %s\n", __func__, this->name);

    if(down_interruptible(&(phy->mutex))){
        return -EINTR;
    }

    if(cmd->autoneg == AUTONEG_DISABLE){
        dummy_adv = 0;

        switch(cmd->speed){
        case SPEED_10:
            if(cmd->duplex == DUPLEX_FULL){
                dummy_adv = ADVERTISED_10baseT_Full;
            }else{
                dummy_adv = ADVERTISED_10baseT_Half;
            }
            break;
        case SPEED_100:
            if(cmd->duplex == DUPLEX_FULL){
                dummy_adv = ADVERTISED_100baseT_Full;
            }else{
                dummy_adv = ADVERTISED_100baseT_Half;
            }
            break;
        case SPEED_1000:
            if(cmd->duplex == DUPLEX_FULL){
                dummy_adv = ADVERTISED_1000baseT_Full;
            }else{
                dummy_adv = ADVERTISED_1000baseT_Half;
            }
            break;
        default:
            AVMNET_ERR("[%s] PHY %s: received bogus ethtool_cmd.\n", __func__, this->name);
            up(&(phy->mutex));
            return -EINVAL;
        }
    }else{
        dummy_adv = cmd->autoneg;
    }

    if((phy->config.supported & dummy_adv) != dummy_adv){
        AVMNET_ERR("[%s] PHY %s: requested mode(s) not supported.\n", __func__, this->name);
        up(&(phy->mutex));
        return -EINVAL;
    }

#if 0
    // refuse GBit settings if not in GBit mode
    if(!(phy->config.supported & AVMNET_CONFIG_FLAG_PHY_GBIT)){
        if(cmd->autoneg == AUTONEG_ENABLE){
            if(cmd->advertising & (ADVERTISED_1000baseT_Half | ADVERTISED_1000baseT_Full)){
                AVMNET_ERR("[%s] Asked to set GBit-Adv for FE Phy!\n", __func__);
                dump_stack();
                up(&(phy->mutex));
                return -EINVAL;
            }
        }else if(cmd->speed == SPEED_1000){
            AVMNET_ERR("[%s] Asked to set GBit speed on FE Phy!\n", __func__);
            dump_stack();
            up(&(phy->mutex));
            return -EINVAL;
        }
    }
#endif

    if(cmd->autoneg == AUTONEG_ENABLE){
        // pause parameter advertising is controlled via ethtool_set_pauseparam()
        cmd->advertising &= ~(ADVERTISED_Pause | ADVERTISED_Asym_Pause);
        cmd->advertising |= (phy->config.advertise & (ADVERTISED_Pause | ADVERTISED_Asym_Pause));
        phy->config.advertise = cmd->advertising;
        phy->config.autoneg = AUTONEG_ENABLE;
    }else{
        phy->config.speed = cmd->speed;
        phy->config.duplex = cmd->duplex;
        phy->config.autoneg = AUTONEG_DISABLE;
    }

    result = set_config(this);

    up(&(phy->mutex));

    return result;
}
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void avmnet_ar8033_ethtool_get_pauseparam(avmnet_module_t *this,
        struct ethtool_pauseparam *param)
{
    struct athphy_context *phy = (struct athphy_context *) this->priv;
    avmnet_linkstatus_t status;

    status = get_status(this);

    if(phy->config.autoneg){
        param->autoneg = AUTONEG_ENABLE;
    }else{
        param->autoneg = AUTONEG_DISABLE;
    }

    if(status.Details.flowcontrol & AVMNET_LINKSTATUS_FC_RX){
        param->rx_pause = 1;
    }

    if(status.Details.flowcontrol & AVMNET_LINKSTATUS_FC_TX){
        param->tx_pause = 1;
    }
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8033_ethtool_set_pauseparam(avmnet_module_t *this,
        struct ethtool_pauseparam *param)
{
    struct athphy_context *phy = (struct athphy_context *) this->priv;
    u32 new_adv;
    enum avmnet_pause_setup orig_pause_setup;

    if(!param->autoneg && (param->rx_pause || param->tx_pause)){
        AVMNET_ERR("[%s] ethtool request to force set flow control parameters, but can only be auto-negotiated\n", __func__);
        return -EINVAL;
    }

    down(&(phy->mutex));

    orig_pause_setup = phy->config.pause_setup;
    new_adv = phy->config.advertise;
    new_adv &= ~(ADVERTISED_Asym_Pause | ADVERTISED_Pause);
    if(param->autoneg){

        phy->config.pause_setup = avmnet_pause_none;

        // does our MAC support ASYM_PAUSE?
        if (this->parent->initdata.mac.flags & AVMNET_CONFIG_FLAG_MAC_SUPPORT_FC_ASYM){

            if(!param->rx_pause && param->tx_pause){
                phy->config.pause_setup = avmnet_pause_tx;
            }

            if(param->rx_pause && !param->tx_pause){
                phy->config.pause_setup = avmnet_pause_rx;
            }
        }

        // does our MAC support any PAUSE?
        if (this->parent->initdata.mac.flags & (AVMNET_CONFIG_FLAG_MAC_SUPPORT_FC_SYM | AVMNET_CONFIG_FLAG_MAC_SUPPORT_FC_ASYM)){

            if(param->rx_pause && param->tx_pause){
                phy->config.pause_setup = avmnet_pause_rx_tx;
            }
        }

        // calculate our current advertisement
        new_adv |= calc_advertise_pause_param(
            (this->parent->initdata.mac.flags &  AVMNET_CONFIG_FLAG_MAC_SUPPORT_FC_SYM),
            (this->parent->initdata.mac.flags &  AVMNET_CONFIG_FLAG_MAC_SUPPORT_FC_ASYM),
            phy->config.pause_setup );

        if((phy->config.advertise == new_adv) && (orig_pause_setup == phy->config.pause_setup)){
            up(&(phy->mutex));
            return -EINVAL;
        }

        phy->config.advertise = new_adv;
        set_config(this);
    }

    up(&(phy->mutex));

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int set_config(avmnet_module_t *this)
{
    struct athphy_context *phy = (struct athphy_context *) this->priv;
    unsigned int phyCtrl, phyAdv, phyGCtrl = 0;
    /*--- unsigned int dummy; ---*/

    if(!down_trylock(&(phy->mutex))){
        AVMNET_ERR("[%s] Called without MUTEX held!\n", __func__);
        dump_stack();
        up(&(phy->mutex));
    }

#if 0
    // disable SmartSpeed
    MDIO_WRITE(ATHR_PHY_SMARTSPEED, 0);
    MDIO_WRITE(MDIO_CONTROL, MDIO_CONTROL_RESET);

    // disable SmartEEE
    MDIO_WRITE(MDIO_MMD_CTRL, 0x3);
    MDIO_WRITE(MDIO_MMD_ADDR, 0x805D);
    MDIO_WRITE(MDIO_MMD_CTRL, 0x4003);
    dummy = MDIO_READ(MDIO_MMD_ADDR);
    MDIO_WRITE(MDIO_MMD_ADDR, 0x0);
    AVMNET_ERR("[%s] SmartEEE CTRL3: %04x\n", __func__, dummy);

    // disable EEE advertisement
    MDIO_WRITE(MDIO_MMD_CTRL, 0x7);
    MDIO_WRITE(MDIO_MMD_ADDR, 0x3C);
    MDIO_WRITE(MDIO_MMD_CTRL, 0x4007);
    dummy = MDIO_READ(MDIO_MMD_ADDR);
    MDIO_WRITE(MDIO_MMD_ADDR, 0x0);
    AVMNET_ERR("[%s] EEE Adv: %04x\n", __func__, dummy);

    // disable hibernation
    MDIO_WRITE(ATHR_PHY_DEBUG_PORT_ADDRESS, 0xB);
    MDIO_WRITE(ATHR_PHY_DEBUG_PORT_DATA, 0);
#endif

    phyCtrl = 0;

    phyAdv = MDIO_READ(MDIO_AUTONEG_ADV);
    phyAdv &= ~(MDIO_AUTONEG_ADV_SPEED_MASK | MDIO_AUTONEG_ADV_PAUSE_SYM | MDIO_AUTONEG_ADV_PAUSE_ASYM);

    if(phy->config.advertise & ADVERTISED_10baseT_Half){
        phyAdv |= MDIO_AUTONEG_ADV_10BT_HDX;
    }
    if(phy->config.advertise & ADVERTISED_10baseT_Full){
        phyAdv |= MDIO_AUTONEG_ADV_10BT_FDX;
    }
    if(phy->config.advertise & ADVERTISED_100baseT_Half){
        phyAdv |= MDIO_AUTONEG_ADV_100BT_HDX;
    }
    if(phy->config.advertise & ADVERTISED_100baseT_Full){
        phyAdv |= MDIO_AUTONEG_ADV_100BT_FDX;
    }
    if(phy->config.advertise & ADVERTISED_Asym_Pause){
        phyAdv |= MDIO_AUTONEG_ADV_PAUSE_ASYM;
    }
    if(phy->config.advertise & ADVERTISED_Pause){
        phyAdv |= MDIO_AUTONEG_ADV_PAUSE_SYM;
    }

    if(phy->config.autoneg == AUTONEG_ENABLE){
        phyCtrl |= MDIO_CONTROL_AUTONEG_ENABLE;
    }

    phyGCtrl = MDIO_READ(MDIO_MASTER_SLAVE_CTRL);
    phyGCtrl &= ~(MDIO_MASTER_SLAVE_CTRL_ADV_FD | MDIO_MASTER_SLAVE_CTRL_ADV_HD);

    if(phy->gbit){
        if(phy->config.advertise & ADVERTISED_1000baseT_Full){
            phyGCtrl |= MDIO_MASTER_SLAVE_CTRL_ADV_FD;
        }
    } 

    switch(phy->config.speed){
    case SPEED_10:
        // bits already cleared
        break;
    case SPEED_100:
        phyCtrl |= MDIO_CONTROL_SPEEDSEL_LSB;
        break;
    case SPEED_1000:
        if(phy->gbit){
            phyCtrl |= MDIO_CONTROL_SPEEDSEL_MSB;
        }else{
            AVMNET_ERR("[%s] PHY %s not supporting GBit speed\n", __func__, this->name);
        }
        break;
    default:
        // default to 10MBit?
        break;
    }

    if(phy->config.duplex){
        phyCtrl |= MDIO_CONTROL_FULLDUPLEX;
    }

    if(phy->config.autoneg == AUTONEG_ENABLE){
        phyCtrl |= MDIO_CONTROL_AUTONEG_RESTART;
    }

    AVMNET_TRC("[%s] module '%s' phyCtrl: 0x%04x phyAdv: 0x%04x phyGCtrl: 0x%04x\n",
            __func__, this->name, phyCtrl, phyAdv, phyGCtrl);

    MDIO_WRITE(MDIO_AUTONEG_ADV, phyAdv);
    MDIO_WRITE(MDIO_CONTROL, phyCtrl);

    MDIO_WRITE(MDIO_MASTER_SLAVE_CTRL, phyGCtrl);   /*--- set or clear 1GBit Advertisement ---*/

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
u32 avmnet_ar8033_ethtool_get_link(avmnet_module_t *this)
{
    avmnet_linkstatus_t status;

    status = get_status(this);

    return status.Details.link;
}
/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
unsigned int avmnet_ar8033_reg_read(avmnet_module_t *this, unsigned int addr, unsigned int reg)
{
    return this->parent->reg_read(this->parent, addr, reg);
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8033_reg_write(avmnet_module_t *this, unsigned int addr, unsigned int reg, unsigned int val)
{
    return this->parent->reg_write(this->parent, addr, reg, val);
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8033_lock(avmnet_module_t *this)
{
    struct athphy_context *ctx;
    ctx = (struct athphy_context *) this->priv;

    return down_interruptible(&(ctx->mutex));
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void avmnet_ar8033_unlock(avmnet_module_t *this)
{
    struct athphy_context *ctx;
    ctx = (struct athphy_context *) this->priv;

    up(&(ctx->mutex));
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void avmnet_ar8033_status_changed(avmnet_module_t *this, avmnet_module_t *caller)
{
    int i;
    // struct athphy_context *my_ctx = (struct athphy_context *) this->priv;
    avmnet_module_t *child = NULL;

    /*
     * find out which of our progeny demands our attention
     */
    for(i = 0; i < this->num_children; ++i){
        if(this->children[i] == caller){
            child = this->children[i];
            break;
        }
    }

    if(child == NULL){
        AVMNET_ERR("[%s] module %s: received status_changed from unknown module.\n", __func__, this->name);
        return;
    }

    /*
     * handle status change
     */

    return;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8033_suspend(avmnet_module_t *this __attribute__ ((unused)), avmnet_module_t *caller __attribute__ ((unused)))
{
    // TODO

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8033_resume(avmnet_module_t *this __attribute__ ((unused)), avmnet_module_t *caller __attribute__ ((unused)))
{
    // TODO

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
int avmnet_ar8033_reinit(avmnet_module_t *this __attribute__ ((unused)))
{
    // TODO

    return 0;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
static int phy_reg_show(struct seq_file *seq, void *data __attribute__ ((unused)))
{
    avmnet_module_t *this;
    struct athphy_context *phy;
    int mdio_reg, mdio_addr, result;
    unsigned int reg_val;

    this = (avmnet_module_t *) seq->private;
    phy = (struct athphy_context *) this->priv;
    result = 0;

    mdio_addr = this->initdata.phy.mdio_addr;
    seq_printf(seq, "----------------------------------------\n");
    seq_printf(seq, "dumping regs for phy %s [mdio:%#x]\n", this->name, mdio_addr);

    for(mdio_reg = 0; mdio_reg < 0x20; ++mdio_reg){
        if(down_interruptible(&(phy->mutex))){
            AVMNET_INFO("[%s] Interrupted while waiting for lock.\n", __func__);
            result = -EINTR;
            break;
        }
        reg_val = MDIO_READ(mdio_reg);
        /*
         * link state is latched low in status reg, so . Make sure we don't miss a lost link
         * indicator.
         */
        if(mdio_reg == MDIO_STATUS && phy->state == LINK_UP && !(reg_val & MDIO_STATUS_LINK)){
            phy->state = LINK_LOST;
            phy->lost_time = jiffies;
        }
        up(&(phy->mutex));

        seq_printf(seq, "    reg[%#x]: %#x\n", mdio_reg, reg_val);
    }
    seq_printf(seq, "----------------------------------------\n");

    return result;
}

static int phy_reg_open(struct inode *inode, struct file *file)
{
    avmnet_module_t *this;

    this = (avmnet_module_t *) PDE(inode)->data;

    return single_open(file, phy_reg_show, this);
}

static ssize_t phy_reg_write(struct file *filp, const char __user *buff, size_t len,
        loff_t *offset __attribute__ ((unused)))
{
    avmnet_module_t *this;
    struct athphy_context *phy;
    struct seq_file *seq;
    char mybuff[128];
    unsigned int reg, value;

    if(len >= 128){
        return -ENOMEM;
    }

    seq = (struct seq_file *) filp->private_data;
    this = (avmnet_module_t *) seq->private;
    phy = (struct athphy_context *) this->priv;

    copy_from_user(&mybuff[0], buff, len);
    mybuff[len] = 0;

    if(sscanf(mybuff, "%i %i", &reg, &value) == 2){
        if(down_interruptible(&(phy->mutex))){
            AVMNET_INFO("[%s] Interrupted while waiting for lock.\n", __func__);
            return -EINTR;
        }

        MDIO_WRITE(reg, value);
        up(&(phy->mutex));
    }

    return len;
}

static void reset_phy(avmnet_module_t *this, enum rst_type type)
{
    struct athphy_context *phy = (struct athphy_context *) this->priv;

    if((this->initdata.phy.flags & AVMNET_CONFIG_FLAG_RESET) == 0){
        AVMNET_ERR("[%s] Reset not configured for module %s\n", __func__, this->name);
        return;
    }

    if(!down_trylock(&(phy->mutex))){
        AVMNET_ERR("[%s] Called without MUTEX held!\n", __func__);
        dump_stack();
        up(&(phy->mutex));
    }

    if(type == rst_on || type == rst_pulse){
        ath_avm_gpio_out_bit(phy->reset, 0);
        mdelay(100);
    }

    if(type == rst_off || type == rst_pulse){
        ath_avm_gpio_out_bit(phy->reset, 1);
        mdelay(100);
    }
}
